import psyco
import math, geometry, random
from skeleton import Skeleton   
import pygame
from geometry import to_radians, to_degrees
import drawing
import centroid_spring
import names
import vector
from vector import *

import numpy, pylab

psyco.full()

class Model:
        def __init__(self, steps, prob):
            self.step = 0
            self.steps = steps
            self.prob = prob
            
            
        def next(self):
            self.step = self.step+1
            if self.step%self.steps==0:
                return random.random()<self.prob
            else:
                return False    
                
        def __call__(self, *args, **kwargs):
            return self.next()

def quantized_exponential_model(steps, prob):               
    return Model(steps, prob)
    
    
def exponential_model(prob):
    return quantized_exponential_model(1, prob)
                
def quantized_model(steps):
    return quantized_exponential_model(steps, 1)
    
       
def gaussian_distribution(var):
    return lambda: random.gauss(0, var)
       
def uniform_distribution(width):
    return lambda: random.randrange(-width, width)
    
def uniform_range_distribution(start, end):
    return lambda: random.randrange(start, end)
    
def constant_distribution(x):
    return lambda: x
        

class NavStats:
    def __init__(self):
        self.log = []
        
    def update(self, wanderers, centroid_pos, max_distance, elapsed, converged):
        # record the statistics that change
        row = {}
        row['elapsed'] = elapsed
        row['centroid'] = centroid_pos
        row['max_distance'] = max_distance
        row['converged'] = converged
        
        wanderers = []
        for wanderer in wanderers:
            subrow = {}
            wanderers['pos'] = wanderer.pos
            wanderers['offset_pos'] = wanderer.offset_pos
            wanderers['halting'] = wanderer.halting
            wanderers['heading'] = wanderer.heading
            wanderers['offset_angle'] = wanderer.offset_anlge
            wanderers['distance'] = wanderer.distance
            wanderers['true_target_distance'] = wanderer.true_target_distance
            wanderers.append(subrow)
        row['wanderers'] = wanderers
        self.log.append(row)
        
            
        

class Wanderer:
    def __init__(self, pos, step_size=1, init_heading = 0, angle_uncertainty=5, position_uncertainty = 50, angle_width = 60, mode="worst", offset_update_prob=0.01, angle_update_prob=0.01, stop_radius=100, halt_time=5, name="Wanderer"):
        self.name = name
        self.pos = pos
        self.step_size = step_size
        self.random_offset = (0,0)
        self.angle_update_prob = angle_update_prob
        self.offset_update = offset_update_prob
        self.heading = geometry.to_radians(init_heading)
        self.angle_uncertainty = to_radians(angle_uncertainty)
        self.position_uncertainty = position_uncertainty
        self.angle_width = to_radians(angle_width/2)
        self.true_target_distance = 0
        self.mode = mode
        self.distance = 0
        self.history = []
        self.stopped = False
        self.stop_radius = stop_radius
        self.halting = 0
        self.halt_time = halt_time
        
        
    def step_towards(self, target, obstacles = None):
    
            
        # update random gps error
        if random.random()<self.offset_update:
            self.random_offset = (random.gauss(0, self.position_uncertainty),random.gauss(0, self.position_uncertainty))
        
        # get new offseted position
        x = self.pos[0] + self.random_offset[0]
        y = self.pos[1] + self.random_offset[1]
        
                    
        
        self.offset_pos = (x,y)
        
        self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
        
        
        # compute angle to target
        dx = target[0] - x
        dy = target[1] - y
        ang = math.atan2(-dy, dx)
        l = math.sqrt(dx*dx+dy*dy)
        
        self.target_distance = l
        self.true_target_distance = vdistance(self.pos, target)
        
        angle = geometry.shortest_angle(self.offset_angle, ang)
        
        
        # if it's time to check...
        if random.random()<self.angle_update_prob:
            self.wall_ended = False
            self.halting = self.halt_time
            # can we detect this angle?
            if abs(angle)>self.angle_width:
                if self.mode=="worst":
                    delta = (abs(angle)-self.angle_width) * (angle/abs(angle))
                    self.heading = self.heading + delta
                    self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
                if self.mode=="random":
                    while abs(angle)>self.angle_width:
                        self.heading = random.random()*math.pi*2 - math.pi
                        self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
                        angle = geometry.shortest_angle(self.offset_angle, ang)
                        
                        
                
        self.history.append(self.pos)
                
            
        # compute step velocity
        move_x = math.cos(self.heading)*self.step_size
        move_y = -math.sin(self.heading)*self.step_size
        next_pos = (self.pos[0]+move_x, self.pos[1]+move_y)
        
        #check obstacles
        if obstacles:
            for obstacle in obstacles:
                direction = obstacle.intersect(self.pos, next_pos)
                # ok we hit something, change our tack...
                if direction!=None:
                    self.heading = math.atan2(-direction[1], direction[0])
                    return
                    
                                                        
        
        # step forward
        if l<self.stop_radius:
            self.stopped = True            
        else:
            if self.halting>0:
                self.halting = self.halting-1
            else:
                
                self.distance = self.distance + self.step_size
                self.pos = next_pos
                
                
        
        
        
        
def format_time(seconds):
    if seconds<60:
        return "%d seconds" % seconds
    if seconds>60 and seconds<3600:
        return "%d minutes %d seconds" % (seconds//60, seconds%60)  
    if seconds>=3600:
        return "%d hours %d minutes %d seconds" % (seconds//3600, (seconds//60)%60, seconds%60)
  
 
 
 
class ObstaclePolygon:
    
    # represent an obstacle as a polygon
    def __init__(self, pts):
        self.pts = pts
        self.lines = []
        old = None
        for pt in pts:
            if old!=None:
                self.lines.append([pt,old])
            old = pt           
        self.lines.append([pts[0], pts[-1]])
        
            
    def inside(self, pt):
        # return true if pt is inside the polygon.
        # assuming that it is convex, of course...
        # slow
        sum = reduce(vadd,self.pts)
        centroid = vscale(sum, 1.0/len(self.pts))
        for i in range(len(self.pts)-1):
            if geometry.inside_triangle(pt,[centroid, self.pts[i], self.pts[i+1]]):
                return True
        return False
        
        
        
        
        
    def intersect(self, start, end):
        
        # intersect the line from start to end with each line in the geometry
        # then find the direction with the biggest dot product
        # this is the direction closest to the direction we are already heading in        
        for line in self.lines:           
            # check interesections (both ways, lines are solid from both sides!)
            
            aside = geometry.line_intersect(start,end, line[0], line[1])
            bside = geometry.line_intersect(end,start, line[0], line[1])
            
            # if intersect, compute dot product, return best fitting direction
            if aside or bside:
                adir = vnormalize(vsub(line[0], line[1]))
                bdir = vnormalize(vsub(line[1], line[0]))
                adp = vdotproduct(adir, vsub(start, end))
                bdp = vdotproduct(bdir, vsub(start,end))
                if adp<bdp:
                    return adir
                else:
                    return bdir                                                               
        return None
                
                
        
def create_report(world, logs, iterations):
    
    
    end_times = numpy.array([log[-1]['elapsed'] for log in logs])
    init_distances = numpy.array([log[0]['max_distance'] for log in logs])
    
    # raw times    
    pylab.figure(1)
    pylab.title("End times for convergence (mean=%.1f, std=%.1f)" % (numpy.mean(end_times), numpy.std(end_times)))
    pylab.ylabel("Frequency")
    pylab.xlabel("Time (s)")
    pylab.hist(end_times, 20)
    pylab.axis([0,3600,0,iterations/3])
    pylab.savefig("convtime.eps")
    
    # speed (max_distance/time)
    pylab.figure(2)
    speed = init_distances/(end_times+10)
    pylab.title("Overall Speed (mean=%.2f, std=%.2f)" % (numpy.mean(speed), numpy.std(speed)))
    pylab.ylabel("Frequency")
    pylab.xlabel("Speed (m/s)")        
    pylab.hist(speed, 20)
    pylab.axis([0,1.0,0,iterations/3])
    pylab.savefig("convspeed.eps")
    
    
    pylab.figure(3)
    pylab.hold(True)
    for log in logs:
        max_distance = numpy.array([v['max_distance'] for v in log])
        pylab.plot(max_distance, 'b')
    pylab.title("Maximum distance of wanderer to target")
    pylab.ylabel("Distance (m)")
    pylab.xlabel("Time (s)")        
    pylab.savefig("walktime.eps")
    
    
    pylab.show()
 
class World:

    def choose_position(self, field_size):
            pos = (random.randint(-field_size/2,field_size/2), random.randint(-field_size/2,field_size/2))

            
            for obstacle in self.obstacles:
                # recurse if still inside
                if obstacle.inside(pos):
                    return self.choose_position(field_size)
            return pos
                
                
                
    def reset_wanderers(self):
        self.nav_log = NavStats()
        self.elapsed = 0
        self.converged = False
        self.wanderers = []
        # create the wanderers
        for i in range(self.n_wanderers):        
            pos= self.choose_position(self.field_size)
            self.wanderers.append(Wanderer(pos=pos, stop_radius=self.stop_radius, name=names.random_name(), **self.wanderer_args))           
            
        
        # step them once to initialize all variables
        for wanderer in self.wanderers:            
            wanderer.step_towards((0,0))
            
        # if spring, start at current average position
        if self.sprung and self.n_wanderers>1:            
            self.compute_geometric_centroid()
            self.spring = centroid_spring.Centroid(pos=self.target, **spring_args)
            
        
    def __init__(self,  spring_args = {}, n_wanderers=5,  field_size=300, scale=1.0, wanderer_args={}, stop_radius = 50, sprung = False):
        self.target = (0,0)
        
        self.spring_args = spring_args
        self.field_size = field_size
        
        self.n_wanderers = n_wanderers
        self.wanderer_args = wanderer_args
        self.stop_radius = stop_radius
        self.sprung = sprung
        # create the obstacles
        self.obstacles = [] #[ObstaclePolygon([(100,10), (10,10), (10,140), (200,120)])]

        
        
    def run(self,mode="interactive", bailout=3600):    
        self.reset_wanderers()        
        if mode=="interactive":
            self.skeleton = Skeleton(tick_fn = self.tick, draw_fn = self.draw)
            font_name = pygame.font.match_font('bitstreamverasansmono', 'verdana', 'sans')
            self.skeleton.default_font = pygame.font.Font(font_name, 8)
        
            self.skeleton.main_loop()
        else:            
            while not self.converged and self.elapsed<bailout:
                self.tick(1)
        
                
    def batch(self, iterations):
        logs = []
        for i in range(iterations):
            self.run(mode="batch")
            logs.append(self.nav_log.log)
            print "Iteration %d of %d" % (i, iterations)
                    
        create_report(self, logs, iterations)
            
        
        
    def compute_geometric_centroid(self):
        # compute the average position of all wanderers
            av_x = 0
            av_y = 0
            for wanderer in self.wanderers:
                av_x = av_x + wanderer.offset_pos[0]
                av_y = av_y + wanderer.offset_pos[1]
            self.target = (av_x/float(len(self.wanderers)), av_y/float(len(self.wanderers)))
        
        
    def tick(self, dt):   
        # check if converged on target
        max_distance = max([wanderer.true_target_distance for wanderer in self.wanderers])        
        if max_distance<self.stop_radius:
            self.converged=True        
        steps = 2             
        if not self.converged:
            # update time
            self.elapsed = self.elapsed+steps
            
            # move wanderers
            for i in range(steps):
                for wanderer in self.wanderers:
                    wanderer.step_towards(self.target, obstacles = self.obstacles)
                    

            # update centroid
            if self.n_wanderers>=2:
                # spring or simple geometric centroid
                if self.sprung:
                    positions = [wanderer.pos for wanderer in self.wanderers]
                    self.spring.set_targets(positions)
                    self.spring.integrate(0.1)
                    self.target = self.spring.x
                else:
                    self.compute_geometric_centroid()
            else:
                self.target = (0,0)
                    
        # update the logs
        self.nav_log.update(self.wanderers, self.target, max_distance, self.elapsed, self.converged)
            
        

        
        
        
    
    def draw_wanderers(self, surface):
        
        for wanderer in self.wanderers:
            x = wanderer.pos[0]+self.skeleton.centre_x
            y = wanderer.pos[1]+self.skeleton.centre_y
            drawing.aacircle(surface, (255,190,100), (x,y), 3)            
                                   
            
            nx = wanderer.offset_pos[0]+self.skeleton.centre_x
            ny = wanderer.offset_pos[1]+self.skeleton.centre_y            
            #drawing.aacircle(surface, (255,100,100), (nx,ny), 5)
            pygame.draw.aaline(surface, (255*0.5,190*0.5,100*0.5), (nx,ny), (x,y))
            
            
            l = 50
            dx = math.cos(wanderer.heading-wanderer.angle_width) *l
            dy = -math.sin(wanderer.heading-wanderer.angle_width) *l
            pygame.draw.aaline(surface, (255,150,50), (nx,ny), (nx+dx, ny+dy))
            
            
            self.skeleton.text(surface, wanderer.name, (255,190,100), (x,y))
          
            
            dx = math.cos(wanderer.heading+wanderer.angle_width) * l            
            dy = -math.sin(wanderer.heading+wanderer.angle_width) * l
            pygame.draw.aaline(surface, (255,150,50), (nx,ny), (nx+dx, ny+dy))
            
            drawing.aacircle(surface, (255,190,100), (nx,ny), 40, wanderer.heading-wanderer.angle_width,
            wanderer.heading+wanderer.angle_width)            
            
            #for i in range(0, len(wanderer.history),10):
            #    p = wanderer.history[i]
            #    pygame.draw.circle(surface, (255,200,200), (p[0]+self.skeleton.centre_x,p[1]+self.skeleton.centre_y), 2)
            
            
    def draw_obstacles(self, surface):
        for obstacle in self.obstacles:
            pts = []
            for pt in obstacle.pts:
                pts.append((pt[0]+self.skeleton.centre_x, pt[1]+self.skeleton.centre_y))
            pygame.draw.aalines(surface, (100,40,0), True, pts)
    
    def draw(self, surface):
        surface.fill((0,0,0))
        drawing.aacircle(surface, (255,255,255), (self.target[0]+self.skeleton.centre_x, self.target[1]+self.skeleton.centre_y), self.stop_radius)
        drawing.aacircle(surface, (255,255,255), (self.target[0]+self.skeleton.centre_x, self.target[1]+self.skeleton.centre_y), 2)
        
        
        max_distance = max([wanderer.true_target_distance for wanderer in self.wanderers])
        
        drawing.aacircle(surface, (48,48,48), (self.target[0]+self.skeleton.centre_x, self.target[1]+self.skeleton.centre_y), max_distance)
        
        self.skeleton.text(surface, format_time(self.elapsed), (255,255,255), (20, 20))
        self.draw_wanderers(surface)
        self.draw_obstacles(surface)
        
        
        
            
            
        
        
        
        
wanderer_args = { 
  "init_heading":random.randrange(0,360), 
  "position_uncertainty":15, 
  "angle_width":60, 
  "angle_uncertainty":7, 
  "offset_update_prob":1/60.0, 
  "angle_update_prob":1/60.0, 
  "mode":"worst",
  "halt_time":25,
  "step_size":0.56}
  
spring_args = {
"damping":1-1e-2,
"friction":0,
"k":1e-2}

world_args = {
"sprung":True, 
"stop_radius":20,
"field_size":500,
"n_wanderers":1,
}

w = World(wanderer_args = wanderer_args, spring_args=spring_args, **world_args)
w.batch(100)


#w.run()

