import cPickle
import os
from Tkinter import *
import tkSimpleDialog, tkMessageBox
from simulator import World
import random
import Tix
import polyeditor


class NavGUI:


    def __init__(self):
        self.load_previous_settings()
        self.init_gui()
        
        
        
    def edit_model(self):
        filename = self.model_tk.get()
        
        try:
            f = open(filename+".obs", "rb")
        except:
            print "Could not edit model..."

        polys = cPickle.load(f)
        f.close()        
        
        field_size = self.field_size_tk.get()
        p = polyeditor.Polyeditor(krange=(-field_size/2, field_size/2), polys=polys)
        
        p.main_loop()
        
        if tkMessageBox.askyesno("Save", "Save changes?"):
            f = open(filename+".obs", "wb")
            cPickle.dump(p.polys,f)
            f.close()           
        
        
        
        
        
    def new_model(self):
    
        field_size = self.field_size_tk.get()
        p = polyeditor.Polyeditor(krange=(-field_size/2, field_size/2))
        p.main_loop()
        
        filename = tkSimpleDialog.askstring("Obstacle name?", "")
        while not filename:
            filename = tkSimpleDialog.askstring("Obstacle name?", "")
            
        f = open(filename+".obs", "wb")
        cPickle.dump(p.polys, f)
        f.close()
        
        self.obstacle_combo.insert(0, filename)
        self.model_tk.set(filename)
        
        
    def add_entry(self,frame, label, init_value, varname, vtype=IntVar):
        setattr(self, varname, vtype())
        getattr(self, varname).set(init_value)
                
        
        lt = Tix.LabelEntry(frame, label=label)
        lt.pack(fill=X)
        lt.entry['textvariable'] =  getattr(self, varname)

    def add_separator(self, frame):
        separator = Frame(frame, height=2, bd=1, relief=SUNKEN)
        separator.pack(fill=X, padx=5, pady=5)
        
        
    def get_model_list(self):
        models = ["blank"]
        file_list = os.listdir(".")
        for file in file_list:
            (path, file) = os.path.split(file)
            (file, extension) = os.path.splitext(file)
            if extension==".obs":
                models.append(file)
        return models
        

    def init_gui(self):
        root = Tix.Tk()                
        frame = Frame(root)
        frame.pack()        
        
        
            
        wanderer_frame = Frame(frame)
        self.add_entry(wanderer_frame, "Number of agents", self.world_args["n_wanderers"], "wanderers_tk")
        self.add_entry(wanderer_frame, "Field size (m)", self.world_args["field_size"], "field_size_tk")
        self.add_entry(wanderer_frame, "Target radius (m)", self.world_args["stop_radius"], "stop_radius_tk")
        
        self.fixed_pos_tk = BooleanVar()
        self.fixed_pos_tk.set(self.world_args.get("fixed", False))
        
        Checkbutton(wanderer_frame, text="Fixed positions", variable=self.fixed_pos_tk, indicatoron=0).pack(side="bottom")
        
        
        
        self.model_tk = StringVar()
        obstacle_frame = Frame(wanderer_frame)
        combo = Tix.ComboBox(obstacle_frame, label="Obstacle Model", variable = self.model_tk, value=self.world_args["model"], dropdown=True)
        combo.pack(side="left")
        entry = combo.subwidget_list['entry']
        entry.config(state='readonly')

        obstacle_models = self.get_model_list()
        for obstacle in obstacle_models:
            combo.insert(0,obstacle)
        self.obstacle_combo = combo
        Button(obstacle_frame, text="Edit...", command=self.edit_model).pack(side="left")
        Button(obstacle_frame, text="New...", command=self.new_model).pack(side="left")
        obstacle_frame.pack(fill=X)
       
        
        
        
        # stopping mode
        mode_frame = Frame(wanderer_frame)
        self.stop_mode_tk = StringVar()
        self.stop_mode_tk.set(self.world_args["stop_mode"])
        Label(mode_frame, text="Stopping mode").pack(side="left")
        Radiobutton(mode_frame, text="Line of sight", variable=self.stop_mode_tk, value="los", indicatoron=0).pack(side="left")
        Radiobutton(mode_frame, text="Simple radius", variable=self.stop_mode_tk, value="simple", indicatoron=0).pack(side="left")
        mode_frame.pack(fill=X)
        
        
        # spring params
        spring_frame = Frame(wanderer_frame)
        self.spring_mode_tk = BooleanVar()
        self.spring_mode_tk.set(self.world_args["sprung"])
        Label(spring_frame, text="Centroid mode").pack(side="left")
        Radiobutton(spring_frame, text="Sprung", variable=self.spring_mode_tk, value=True, indicatoron=0).pack(side="left")
        Radiobutton(spring_frame, text="Simple mean", variable=self.spring_mode_tk, value=False, indicatoron=0).pack(side="left")
        spring_frame.pack(fill=X)
        
        wanderer_frame.pack(fill=X)
        self.add_separator(frame)
        agent_frame = Frame(frame)
        self.add_entry(agent_frame, "GPS uncertainty (std. dev. m)", self.wanderer_args["position_uncertainty"], "position_uncertainty_tk", DoubleVar)
        
        self.add_entry(agent_frame, "Angle uncertainty (std. dev. degrees)", self.wanderer_args["angle_uncertainty"], "angle_uncertainty_tk", DoubleVar)
        self.add_entry(agent_frame, "Fbk. Angle width", self.wanderer_args["angle_width"], "angle_width_tk")
        self.add_entry(agent_frame, "Walking rate (m/s)", self.wanderer_args["step_size"], "step_size_tk", DoubleVar)
        self.add_entry(agent_frame, "Scanning time (s)", self.wanderer_args["halt_time"], "halt_time_tk", DoubleVar)
        
        self.add_entry(agent_frame, "Step wandering variation (m)", self.wanderer_args["wander_variation"], "wander_variation_tk", DoubleVar)
        
        
        angle_recip = 1.0/self.wanderer_args["angle_update_prob"]
        offset_recip = 1.0/self.wanderer_args["offset_update_prob"]
        
        self.add_entry(agent_frame, "Heading adjustment period (avg. s)", angle_recip, "angle_update_time_tk", DoubleVar)
        
        self.add_entry(agent_frame, "GPS noise update period (avg. s)", offset_recip, "offset_update_time_tk", DoubleVar)
        
        
        mode_frame = Frame(agent_frame)
        
        self.wander_mode_tk = StringVar()
        self.wander_mode_tk.set(self.wanderer_args["mode"])
        Label(mode_frame, text="Heading mode").pack(side="left")
        Radiobutton(mode_frame, text="Random", variable=self.wander_mode_tk, value="random", indicatoron=0).pack(side="left")
        Radiobutton(mode_frame, text="Worst case", variable=self.wander_mode_tk, value="worst", indicatoron=0).pack(side="left")
        mode_frame.pack(fill=X)
        
        agent_frame.pack(fill=X)
        self.add_separator(frame)
        
        
        
        spring_frame = Frame(frame)
        self.add_entry(spring_frame, "Spring K", self.spring_args["k"], "k_tk", DoubleVar)
        self.add_entry(spring_frame, "Spring damping", self.spring_args["damping"], "damping_tk", DoubleVar)
        self.add_entry(spring_frame, "Spring breaking (m)", self.spring_args["breaking"], "breaking_tk", DoubleVar)
        
        spring_frame.pack(fill=X)
        self.add_separator(frame)
        
        
        self.mode_tk = BooleanVar()
        self.mode_tk.set(self.global_args["interactive"])
        newframe = Frame(frame)
        newframe.pack(fill=X)
        
        Radiobutton(newframe, text="Interactive mode", variable=self.mode_tk, value=True, indicatoron=0).pack(fill=X)
        Radiobutton(newframe, text="Batch mode", variable=self.mode_tk, value=False, indicatoron=0).pack(fill=X)
        
        
        self.add_entry(frame, "Iterations", self.global_args["iterations"], "iterations_tk")
        self.add_separator(frame)        
        
        
        
        
        self.go_button = Button(frame, text="Run", command=self.run)        
        self.go_button.pack()
        
        
        root.mainloop()
        
    
    def save_settings(self):
            f = open("simulatorargs.pickle","wb")
            args_vector = [self.wanderer_args, self.spring_args, self.world_args, self.global_args]
            cPickle.dump(args_vector,f)
            f.close()            
        
        
        
    def map_variables(self):
        self.global_args["interactive"] = self.mode_tk.get()
        self.global_args["iterations"] = self.iterations_tk.get()
        self.world_args["n_wanderers"] = self.wanderers_tk.get()
        self.world_args["field_size"] = self.field_size_tk.get()
        self.world_args["stop_radius"] = self.stop_radius_tk.get()
        self.world_args["stop_mode"] = self.stop_mode_tk.get()
        self.world_args["sprung"] = self.spring_mode_tk.get()
        

        self.world_args["fixed"] = self.fixed_pos_tk.get()
        self.world_args["model"] = self.model_tk.get()
        self.wanderer_args["position_uncertainty"] = self.position_uncertainty_tk.get()
        self.wanderer_args["angle_width"] = self.angle_width_tk.get()
        self.wanderer_args["angle_uncertainty"] = self.angle_uncertainty_tk.get()
        self.wanderer_args["step_size"] = self.step_size_tk.get()
        self.wanderer_args["halt_time"] = self.halt_time_tk.get()        
        self.wanderer_args["mode"] = self.wander_mode_tk.get()
        self.wanderer_args["wander_variation"] = self.wander_variation_tk.get()
        
        self.wanderer_args["offset_update_prob"] = 1.0/self.offset_update_time_tk.get()
        self.wanderer_args["angle_update_prob"] = 1.0/self.angle_update_time_tk.get()
        
        self.spring_args["damping"] = self.damping_tk.get()
        self.spring_args["k"] = self.k_tk.get()
        self.spring_args["breaking"] = self.breaking_tk.get()
        
    
    def run(self):
    
        self.map_variables()
        self.save_settings()
        
        
        self.create_world()
        if self.global_args['interactive']==True:
            self.run_interactive()
        else:
            self.run_batch()
        
    def load_previous_settings(self):    
        args_vector = None
        try:
            f = open("simulatorargs.pickle","rb")
            args_vector = cPickle.load(f)
            f.close()            
        except:
            pass
    
        if args_vector:
            self.wanderer_args, self.spring_args, self.world_args, self.global_args = args_vector
        else:
            print "UO"
            self.wanderer_args = { 
                  "init_heading":random.randrange(0,360), 
                  "position_uncertainty":15, 
                  "angle_width":60, 
                  "angle_uncertainty":7, 
                  "offset_update_prob":1/60.0, 
                  "angle_update_prob":1/60.0, 
                  "mode":"worst",
                  "halt_time":25,
                  "step_size":0.56,
                  "wander_variation":0.2}
                  
            self.spring_args = {
                "damping":1-1e-2,
                "friction":0,
                "k":1e-2,
                "breaking":500}

            self.world_args = {
                "model" : "blank",
                "sprung":True, 
                "stop_radius":20,
                "field_size":500,
                "n_wanderers":1,
                "stop_mode":"los",
                }
                
            self.global_args = {
            "interactive":True,
            "iterations":100}
                   
    
    def create_world(self):
        print self.spring_args
        self.world = World(wanderer_args = self.wanderer_args, spring_args=self.spring_args, **self.world_args)    

        
    def run_interactive(self):
        self.world.run()
        
    def run_batch(self):
        self.world.batch(self.global_args['iterations'])
        
        
n = NavGUI()