import math, geometry, random
from skeleton import Skeleton   
import pygame
from geometry import to_radians, to_degrees
import drawing
import centroid_spring
import names
import vector
from vector import *
import cPickle

from models import *
from wanderer import Wanderer
from report import NavStats, create_report
from obstacle import ObstaclePolygon

import agent_positions
import numpy, pylab



        
        
def format_time(seconds):    
    
    if seconds<60:
        return "%d seconds" % seconds
    if seconds>60 and seconds<3600:
        return "%d minutes %d seconds" % (seconds//60, seconds%60)  
    if seconds>=3600:
        return "%d hours %d minutes %d seconds" % (seconds//3600, (seconds//60)%60, seconds%60)
  
 
 

class World:

    def choose_position(self, index, field_size):
    
        if self.fixed_pos:            
            pos = agent_positions.positions[index]
            
        else:
            pos = (random.randint(-field_size/2,field_size/2), random.randint(-field_size/2,field_size/2))            
            for obstacle in self.obstacles:
                # recurse if still inside
                if obstacle.inside(pos):
                    return self.choose_position(field_size)
        return pos
                
  
    # close the obstacle polygons
    def create_obstacles(self, polys):
        obstacles = []
        for poly in polys:
            if len(poly)>1:
                poly.append(poly[0])
                obstacles.append(ObstaclePolygon(poly))
            
        return obstacles
            
                
    def reset_wanderers(self):
        self.nav_log = NavStats()
        self.elapsed = 0
        self.converged = False
        self.wanderers = []
        # create the wanderers
        for i in range(self.n_wanderers):        
            pos= self.choose_position(i, self.field_size)
            self.wanderers.append(Wanderer(pos=pos, stop_radius=self.stop_radius, name=names.random_name(), **self.wanderer_args))           
            
        
        # step them once to initialize all variables
        for wanderer in self.wanderers:            
            wanderer.step_towards((0,0))
            
        # if spring, start at current average position
        if self.sprung and self.n_wanderers>1:            
            self.compute_geometric_centroid()
            self.spring = centroid_spring.Centroid(pos=self.target, **self.spring_args)
            
        
    def __init__(self,  spring_args = {}, n_wanderers=5,  field_size=300,  wanderer_args={}, stop_radius = 50, sprung = False, stop_mode="los", model="blank", fixed=False):
        self.target = (0,0)
        
        self.fixed_pos = fixed
        
        
        self.spring_args = spring_args
        self.field_size = field_size
        self.stop_mode = stop_mode
        
        self.n_wanderers = n_wanderers
        self.wanderer_args = wanderer_args
        self.stop_radius = stop_radius
        self.sprung = sprung
        
        # create the obstacles (load them from file)
        self.obstacles = []
        if model!="blank":            
            
            f = open(model+".obs", "rb")
            polys = cPickle.load(f)
            f.close()
            
            self.obstacles = self.create_obstacles(polys)
            
                

        
        
    def run(self,mode="interactive", bailout=3600):    
        self.reset_wanderers()        
        if mode=="interactive":
            self.skeleton = Skeleton(tick_fn = self.tick, draw_fn = self.draw)
            font_name = pygame.font.match_font('bitstreamverasansmono', 'verdana', 'sans')
            self.skeleton.default_font = pygame.font.Font(font_name, 10)
        
        
            self.scale = 1.0/((self.field_size*1.5)/float(self.skeleton.h))
            print self.scale
            self.skeleton.main_loop()
        else:            
            while not self.converged and self.elapsed<bailout:
                self.tick(1)
        
                
    def batch(self, iterations):
        logs = []
        for i in range(iterations):
            self.run(mode="batch")
            logs.append(self.nav_log.log)
            print "Iteration %d of %d" % (i, iterations)
                    
        create_report(self, logs, iterations)
            
        
        
    def compute_geometric_centroid(self):
        # compute the average position of all wanderers
            av_x = 0
            av_y = 0
            for wanderer in self.wanderers:
                av_x = av_x + wanderer.offset_pos[0]
                av_y = av_y + wanderer.offset_pos[1]
            self.target = (av_x/float(len(self.wanderers)), av_y/float(len(self.wanderers)))
        
        
    def tick(self, dt):   
        # check if converged on target
        max_distance = max([wanderer.true_target_distance for wanderer in self.wanderers])                
        if self.stop_mode == "simple":
            if max_distance<self.stop_radius:
                self.converged=True        
                
        # line of sight test
        if self.stop_mode == "los":
            
            if max_distance<self.stop_radius:
                los = True
                for w_a in self.wanderers:
                    for w_b in self.wanderers:
                        if w_a!=w_b:
                            if not geometry.line_of_sight(w_a.pos, w_b.pos, [o.pts for o in self.obstacles]):
                                los = False
                if los:
                    self.converged=True
            
                
                
        steps = 2             
        if not self.converged:
            # update time
            self.elapsed = self.elapsed+steps
            
            # move wanderers
            for i in range(steps):
                for wanderer in self.wanderers:
                    wanderer.step_towards(self.target, obstacles = self.obstacles)
                    

            # update centroid
            if self.n_wanderers>=2:
                # spring or simple geometric centroid
                if self.sprung:
                    positions = [wanderer.pos for wanderer in self.wanderers]
                    self.spring.set_targets(positions)
                    self.spring.integrate(0.1)
                    self.target = self.spring.x
                else:
                    self.compute_geometric_centroid()
            else:
                self.target = (0,0)
                    
        # update the logs
        self.nav_log.update(self.wanderers, self.target, max_distance, self.elapsed, self.converged)
            
        
        
        
        
    
    def draw_wanderers(self, surface):
        
        
        
        for wanderer in self.wanderers:
            (x,y) = self.transform(wanderer.pos)
            pygame.draw.circle(surface, (255,190,100), (x,y), 3)            
                                   
            
            (nx,ny) = self.transform(wanderer.offset_pos)            
            pygame.draw.circle(surface, (255,100,100), (nx,ny), 3)
            pygame.draw.aaline(surface, (255*0.5,190*0.5,100*0.5), (nx,ny), (x,y))
            
            
            l = 30
            dx = math.cos(wanderer.heading-wanderer.angle_width) *l
            dy = -math.sin(wanderer.heading-wanderer.angle_width) *l
            pygame.draw.aaline(surface, (255,150,50), (nx,ny), (nx+dx, ny+dy))
            
            
            #self.skeleton.text(surface, wanderer.name, (255,190,100), (x,y))
          
            
            dx = math.cos(wanderer.heading+wanderer.angle_width) * l            
            dy = -math.sin(wanderer.heading+wanderer.angle_width) * l
            pygame.draw.aaline(surface, (255,150,50), (nx,ny), (nx+dx, ny+dy))
            
            drawing.aacircle(surface, (255,190,100), (nx,ny), l*0.6, wanderer.heading-wanderer.angle_width,
            wanderer.heading+wanderer.angle_width)            
            
            pts  = []
            for i in range(0, len(wanderer.history),10):
                pts.append(self.transform(wanderer.history[i]))
            if len(pts)>2:            
                pygame.draw.aalines(surface, (80,80,80), False, pts)
                
            
    def draw_obstacles(self, surface):
        for obstacle in self.obstacles:
            pts = []
            for pt in obstacle.pts:
                pts.append(self.transform(pt))
            pygame.draw.aalines(surface, (100,40,0), True, pts)
    
    
    def transform(self,pt):
        return (pt[0])*self.scale+self.skeleton.centre_x, (pt[1])*self.scale+self.skeleton.centre_y
        
        
        

            
        
    def draw_scale(self, surface):
        x = 0
        s = 0
        while x<self.skeleton.w:
            x = x + 100 * self.scale
            pygame.draw.line(surface, (100,80,80), (x, self.skeleton.h-10),(x, self.skeleton.h-50))
            self.skeleton.text(surface, "%dm" % s, (90,90,90), (x, self.skeleton.h-50))
            s = s + 100
        
    def draw(self, surface):
    
    
        max_distance = max([wanderer.true_target_distance for wanderer in self.wanderers])
        
        
        
        surface.fill((0,0,0))
        
        self.draw_scale(surface)
        
        
        top = self.transform((-self.field_size/2, -self.field_size/2))
        bottom = self.transform((self.field_size/2, self.field_size/2))
        
        r = pygame.Rect(top, vsub(top,bottom))
        
        pygame.draw.rect(surface, (40,40,5), r)
        
    
        
        drawing.aacircle(surface, (255,255,255), self.transform(self.target), self.stop_radius*self.scale)        
        drawing.aacircle(surface, (255,255,255), self.transform(self.target), 2)
        
        
        
        drawing.aacircle(surface, (48,48,48), self.transform(self.target), max_distance*self.scale)
        
        self.skeleton.text(surface, format_time(self.elapsed), (255,255,255), (20, 20))
        self.draw_wanderers(surface)
        self.draw_obstacles(surface)
        
        
        
            
            
        
        
if __name__=="__main__":       
            
    wanderer_args = { 
          "init_heading":random.randrange(0,360), 
          "position_uncertainty":15, 
          "angle_width":60, 
          "angle_uncertainty":7, 
          "offset_update_prob":1/60.0, 
          "angle_update_prob":1/60.0, 
          "mode":"worst",
          "halt_time":25,
          "step_size":0.56}
          
    spring_args = {
    "damping":1-1e-2,
    "friction":0,
    "k":1e-2}

    world_args = {
    "sprung":True, 
    "stop_radius":20,
    "field_size":500,
    "n_wanderers":1,
    }

    w = World(wanderer_args = wanderer_args, spring_args=spring_args, **world_args)    

    w.run()

