import math, geometry, random
from geometry import to_radians, to_degrees
import vector
from vector import *




class Wanderer:
    def __init__(self, pos, step_size=1, init_heading = 0, angle_uncertainty=5, position_uncertainty = 50, angle_width = 60, mode="worst", offset_update_prob=0.01, angle_update_prob=0.01, stop_radius=100, halt_time=5, name="Wanderer", wander_variation=0.2):
        self.name = name
        self.pos = pos
        self.step_size = step_size
        self.random_offset = (0,0)
        self.angle_update_prob = angle_update_prob
        self.offset_update = offset_update_prob
        self.heading = geometry.to_radians(init_heading)
        self.angle_uncertainty = to_radians(angle_uncertainty)
        self.position_uncertainty = position_uncertainty
        self.angle_width = to_radians(angle_width/2)
        self.true_target_distance = 0
        self.mode = mode
        self.distance = 0
        self.history = []
        self.stopped = False
        self.stop_radius = stop_radius
        self.halting = 0
        self.halt_time = halt_time
        self.wander_variation = wander_variation
        self.wall_heading = init_heading
        self.wall_following = False
    


    def step(self, angle, step_size, variation, pos):
        move_x = math.cos(angle)*step_size + random.gauss(0, variation)
        move_y = -math.sin(angle)*step_size + random.gauss(0, variation)
        return (pos[0]+move_x, pos[1]+move_y)
        
        
    def step_towards(self, target, obstacles = None):
    
            
        # update random gps error
        if random.random()<self.offset_update:
            self.random_offset = (random.gauss(0, self.position_uncertainty),random.gauss(0, self.position_uncertainty))
        
        # get new offseted position
        x = self.pos[0] + self.random_offset[0]
        y = self.pos[1] + self.random_offset[1]
        
                    
        
        self.offset_pos = (x,y)
        
        self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
        
        
        # compute angle to target
        dx = target[0] - x
        dy = target[1] - y
        ang = math.atan2(-dy, dx)
        l = math.sqrt(dx*dx+dy*dy)
        
        self.target_distance = l
        self.true_target_distance = vdistance(self.pos, target)
        
        angle = geometry.shortest_angle(self.offset_angle, ang)
        
        
        # if it's time to check...
        if random.random()<self.angle_update_prob and self.halting<=0:
            self.wall_ended = False
            self.halting = self.halt_time
            # can we detect this angle?
            if abs(angle)>self.angle_width:
                if self.mode=="worst":
                    delta = (abs(angle)-self.angle_width) * (angle/abs(angle))
                    self.heading = self.heading + delta
                    self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
                if self.mode=="random":
                    while abs(angle)>self.angle_width:
                        self.heading = random.random()*math.pi*2 - math.pi
                        self.offset_angle = random.gauss(self.heading, self.angle_uncertainty)        
                        angle = geometry.shortest_angle(self.offset_angle, ang)
                        
                        
                
        self.history.append(self.pos)
        
        next_pos = self.step(self.heading, self.step_size, self.wander_variation, self.pos)
            
        
        #check obstacles
        if obstacles:
            for obstacle in obstacles:
                direction = obstacle.intersect(self.pos, next_pos)
                # ok we hit something, change our tack until we find a way forward...
                # nb repetition is required because of the random wander variation
                while direction!=None:
                    angle = math.atan2(-direction[1], direction[0]) + random.random()*0.5-0.25
                    next_pos = self.step(angle, self.step_size, self.wander_variation, self.pos)
                    
                    direction = obstacle.intersect(self.pos, next_pos)
                    self.heading = angle
                    
                                                        
        if l>self.stop_radius/2.0:
            if self.halting>0:
                self.halting = self.halting-1
                
            else:
                
                self.distance = self.distance + self.step_size
                self.pos = next_pos
                
                
        